// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

#include <cmath>
#include <iostream.h>
#include <iomanip.h>
#include <algorithm>
#include "DBase.h"
#include "ArgPack.h"
#include "RGen.h"
#include "Epsilon.h"
#include "RMath.h"

/*###########################################################################*/
DBase::DBase(int table_offset, int type_cnt) :
	type_cnt_(type_cnt),
	table_offset_(table_offset),
	table_(ArgPack::table_cnt_, ArgPack::table_av_.size()),

	table_type_(ArgPack::table_cnt_, type_cnt_, ArgPack::type_av_.size()),

	table_name_(ArgPack::table_name_),
	type_name_(ArgPack::type_name_),
	f_width_(0)
{
	RVector<double> table_stream(ArgPack::table_cnt_);
	MAP(x, table_stream.size())
		table_stream[x] = RGen::gen.flat_range_d(-1.0, 1.0);

	RVector<double> type_stream(type_cnt_);
	MAP(x, type_stream.size())
		type_stream[x] = RGen::gen.flat_range_d(-1.0, 1.0);

/*-------------------------------------*/
	MAP(x, table_name_.size())
		f_width_ = max(f_width_, table_name_[x].size() + 2);

	MAP(x, type_name_.size())
		f_width_ = max(f_width_, type_name_[x].size() + 2);

	MAP2(x, table_.size(), y, table_[x].size()) {
		double jit = RGen::gen.flat_range_d(-1.0, 1.0);

		double deviation =
		  flat_merge(ArgPack::table_jitter_[y], jit, table_stream[x]);

		deviation *= ArgPack::table_mul_[y];

		table_[x][y] = ArgPack::table_av_[y] + deviation;

		if (eps_not_equal_to<double>()(ArgPack::table_round_[y], 0.0)) {
			table_[x][y] =
			  interval_round(table_[x][y], ArgPack::table_round_[y]);
		}
	}

// x: table, y: type, z: attribute
	MAP2(x, table_type_.size(), y, table_type_[x].size()) {
		unsigned imp_cnt = unsigned(rint(
		  ArgPack::entries_per_type_av_ +
		  ArgPack::entries_per_type_mul_ * RGen::gen.flat_range_d(-1.0, 1.0)));

		MAP2(imp, imp_cnt, z, table_type_[x][y].size()) {
			double deviation = flat_merge(ArgPack::table_type_ratio_,
			  table_stream[x], type_stream[y]);

			double jit = RGen::gen.flat_range_d(-1.0, 1.0);

			deviation = flat_merge(ArgPack::type_jitter_[z], jit, deviation);

			deviation *= ArgPack::type_mul_[z];

	 		table_type_[x][y][z].push_back(ArgPack::type_av_[z] + deviation);

			if (eps_not_equal_to<double>()(ArgPack::type_round_[z], 0.0)) {
				table_type_[x][y][z].back() =
				  interval_round(table_type_[x][y][z].back(),
				  ArgPack::type_round_[z]);
			}
		}
	}
}

/*===========================================================================*/
void DBase::print_to(ostream &os) const {
// Print out the tables.
	MAP(x, table_.size()) {
		os << "@" << ArgPack::table_label_ << " " <<
		  (x + table_offset_) << " {\n";

		if (table_name_.size()) {
			os << "# ";

			MAP(y, table_name_.size())
				os << setw(f_width_) << table_name_[y];

			os << "\n  ";

			MAP(y, table_[x].size())
				os << setw(f_width_) << table_[x][y];

			os << "\n";
		}

		if (type_cnt_ && table_name_.size()) {
			os << "\n#----------------------------------------" <<
			  "--------------------------------------\n";
		}

		if (type_cnt_) {
			os << "# " << "type version ";
			MAP(y, type_name_.size())
				os << setw(f_width_) << type_name_[y];

			os << "\n";

// All types, all implementations
			MAP2(y, table_type_[x].size(), implemen, table_type_[x][y][0].size()) {
				os << "  " << setw(strlen("type ")) << y <<
				  setw(strlen("version ")) << implemen;
// implementations, and attributes.
				MAP(z, table_type_[x][y].size())
					os << setw(f_width_) << table_type_[x][y][z][implemen];
				os << "\n";
			}
		}

		os << "}\n\n\n";
	}

	os << "\n";
}


/*===========================================================================*/
void DBase::trans_print_to(ostream &os) const {
// Print out the tables.
	MAP(x, table_.size()) {
		os << "@" << ArgPack::table_label_ << " " <<
		  (x + table_offset_) << " {\n";

		if (table_name_.size()) {
			os << "# ";

			MAP(y, table_name_.size())
				os << setw(f_width_) << table_name_[y];

			os << "\n  ";

			MAP(y, table_[x].size())
				os << setw(f_width_) << table_[x][y];

			os << "\n";
		}

		if (type_cnt_ && table_name_.size()) {
			os << "\n#----------------------------------------" <<
			  "--------------------------------------\n";
		}

		if (type_cnt_) {
			os << "# " << "type ";
			MAP(y, type_name_.size())
				os << setw(f_width_) << type_name_[y];

			os << "\n";

// All types
			MAP(y, table_type_[x].size()) {
				os << "  " << setw(strlen("type ")) << y;
// attributes.
				MAP(z, table_type_[x][y].size()) {
					RASSERT(table_type_[x][y][z].size() == 1);
					os << setw(f_width_) << table_type_[x][y][z].back();
				}
				os << "\n";
			}
		}

		os << "}\n\n\n";
	}

	os << "\n";
}

/*===========================================================================*/
double DBase::flat_merge(double alpha, double a, double b) {
	RASSERT(alpha >= 0.0 && alpha <= 1.0);
	double beta = 1.0 - alpha;

	if (! ArgPack::urv_merge_)
		return alpha * a + beta * b;

	if (beta > alpha) {
		swap(alpha, beta);
		swap(a, b);
	}

	double raw_merge = alpha * a + beta * b;
	double sign = (raw_merge < 0.0) ? -1.0 : 1.0;
	double abs_merge = abs(raw_merge);

	double filt_merge = HUGE_VAL;

	if (abs_merge <= 2.0 * alpha - 1.0) {
		filt_merge = abs_merge / alpha;
	} else {
		filt_merge = abs_merge * abs_merge - 2.0 * abs_merge -
		  4.0 * alpha * alpha + 8.0 * alpha - 3.0;

		filt_merge /= 4.0 * alpha * alpha - 4.0 * alpha;

		filt_merge += 2.0 - 1.0 / alpha;
	}

	double final_merge = sign * filt_merge;
	return final_merge;
}
