// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef EPSILON_H_
#define EPSILON_H_

/*###########################################################################*/
#include <functional>
#include <cmath>
#include <cfloat>
#include "RStd.h"

/*###########################################################################*/
// Only works for floats and doubles.
template <typename T> struct eps_type_info {};

/*===========================================================================*/
template <> struct eps_type_info<float> {
	static const float MIN = FLT_MIN;
	static const float EPSILON = FLT_EPSILON;
};

/*===========================================================================*/
template <> struct eps_type_info<double> {
	static const double MIN = DBL_MIN;
	static const double EPSILON = DBL_EPSILON;
};

/*###########################################################################*/
template <typename T> class eps_not_equal_to;
template <typename T> class eps_greater;
template <typename T> class eps_less_equal;
template <typename T> class eps_greater_equal;
template <typename T> class eps_inc;
template <typename T> class eps_dec;

/*===========================================================================*/
template <typename T>
class eps_equal_to : public binary_function<T, T, bool> {
public:
	eps_equal_to(T slack = 5) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a, T b) const {return equal_to(a, b, min_, eps_);}

private:
	static bool equal_to(T a, T b, T min, T epsilon)

	  {return (abs(a - b) <= min) || abs(a - b) <= abs(a + b) * epsilon;}

		T min_;
		T eps_;

	friend eps_not_equal_to<T>;
};

template <typename T>
bool eps_is_equal_to(T a, T b, T slack = 5)
  {return eps_equal_to<T>(slack)(a, b);}

/*===========================================================================*/
template <typename T>
class eps_less : public binary_function<T, T, bool> {
public:
	eps_less(T slack = 5) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a, T b) const {return less(a, b, min_, eps_);}

private:
	static bool less(T a, T b, T min, T epsilon)
	  {return abs(a - b) > min && a < b - abs(a + b) * epsilon;}

		T min_;
		T eps_;

	friend eps_greater<T>;
	friend eps_less_equal<T>;
	friend eps_greater_equal<T>;
};

template <typename T>
bool eps_is_less(T a, T b, T slack = 5)
  {return eps_less<T>(slack)(a, b);}

/*===========================================================================*/
template <typename T>
class eps_not_equal_to : public binary_function<T, T, bool> {
public:
	eps_not_equal_to(T slack = 5) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a, T b) const
	  {return ! eps_equal_to<T>::equal_to(a, b, min_, eps_);}

private:
		T min_;
		T eps_;
};

template <typename T>
bool eps_is_not_equal_to(T a, T b, T slack = 5)
  {return eps_not_equal_to<T>(slack)(a, b);}

/*===========================================================================*/
template <typename T>
class eps_greater : public binary_function<T, T, bool> {
public:
	eps_greater(T slack = 5) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a, T b) const
	  {return eps_less<T>::less(b, a, min_, eps_);}

private:
		T min_;
		T eps_;
};

template <typename T>
bool eps_is_greater(T a, T b, T slack = 5)
  {return eps_greater<T>(slack)(a, b);}

/*===========================================================================*/
template <typename T>
class eps_less_equal : public binary_function<T, T, bool> {
public:
	eps_less_equal(T slack = 5) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a, T b) const
	  {return ! eps_less<T>::less(b, a, min_, eps_);}

private:
		T min_;
		T eps_;
};

template <typename T>
bool eps_is_less_equal(T a, T b, T slack = 5)
  {return eps_less_equal<T>(slack)(a, b);}

/*===========================================================================*/
template <typename T>
class eps_greater_equal : public binary_function<T, T, bool> {
public:
	eps_greater_equal(T slack = 5) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a, T b) const
	  {return ! eps_less<T>::less(a, b, min_, eps_);}

private:
		T min_;
		T eps_;
};

template <typename T>
bool eps_is_greater_equal(T a, T b, T slack = 5)
  {return eps_greater_equal<T>(slack)(a, b);}

/*===========================================================================*/
template <typename T> class eps_inc : public unary_function<T, T> {
public:
	eps_inc(T slack = 1) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a) const;

private:
		T min_;
		T eps_;
};

/*===========================================================================*/
template <typename T> class eps_dec : public unary_function<T, T> {
public:
	eps_dec(T slack = 1) : min_(slack * eps_type_info<T>::MIN),
	  eps_(slack * eps_type_info<T>::EPSILON) {}

	bool operator()(T a) const;

private:
		T min_;
		T eps_;
};

/*===========================================================================*/
void Epsilon_test();

/*###########################################################################*/
#include "Epsilon.cct"
#endif
