// Copyright 2000 by Robert Dick.
// All rights reserved.

#include <cmath>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include "RGen.h"
#include "RStd.h"
#include "Epsilon.h"

/*##########################################################################*/
#ifdef ROB_DEBUG
static double L_data1[], L_data2[], L_data3[], L_data4[], L_data5[],
  L_data6[], L_data7[], L_data8[], L_data9[], L_data10[];
#endif

/*===========================================================================*/
RGen RGen::gen;

/*===========================================================================*/
RGen::RGen(int seed) :
	u_(97)
{
	set_seed(seed);
}

/*===========================================================================*/
void RGen::set_seed(int seed) {
	int ij = seed / 30082;
	int kl = seed - 30082 * ij;

	int i = ij / 177 % 177 + 2;
	int j = ij % 177 + 2;
	int k = kl / 169 % 178 + 1;
	int l = kl % 169;

	set_seed(i, j, k, l);
}

/*===========================================================================*/
void RGen::set_seed(int i, int j, int k, int l) {
	MAP(ii, u_.size()) {
		float s = 0.0;
		float t = 0.5F;

		for (int jj = 0; jj < 24; jj++) {
			int m = (i * j % 179) * k % 179;

			i = j; j = k; k = m;

			l = (53 * l + 1) % 169;

			if (l * m % 64 >= 32)
				s += t;

			t *= 0.5F;
		}

		u_[ii] = s;
	}

	c_ = 362436.0F / 16777216.0F;
	cd_ = 7654321.0F / 16777216.0F;
	cm_ = 16777213.0F / 16777216.0F;

	ip_ = 96;
	jp_ = 32;
}

/*===========================================================================*/
double RGen::flat01() {
	float uni = u_[ip_] - u_[jp_];
	if (uni < 0.0)
		uni += 1.0F;

	u_[ip_] = uni;
	ip_ ? ip_-- : ip_ = 96;
	jp_ ? jp_-- : jp_ = 96;
	c_ -= cd_;

	if (c_ < 0.0) c_ += cm_;

	uni -= c_;
	if (uni < 0.0) uni += 1.0F;

	RASSERT(uni < 1.0F);

	return uni;
}

/*===========================================================================*/
double RGen::gauss01() {
// Mean 0, variance 1.
		static int recalc=1;
		double in_a, in_b;
		double out_a;
		static double out_b;
		double modifier;
		double compile_b;

		if (recalc) {
			in_a = flat01();
			in_b = flat01();
		
			modifier = sqrt(-2.0 * log(in_a));
			compile_b = 2.0 * PI * in_b;
		
			out_a = modifier * cos(compile_b);
			out_b = modifier * sin(compile_b);

			recalc = 0;
			return out_a;
		}

		recalc = 1;
		return out_b;
}

/*===========================================================================*/
unsigned RGen::operator()(int ceil) {
	return int(ceil * flat01());
}

/*===========================================================================*/
double RGen::flat_range_d(double low, double high) {
	RASSERT(low <= high);
	return (high - low) * flat01() + low;
}

/*===========================================================================*/
long RGen::flat_range_l(long low, long high) {
	RASSERT(low <= high);
	long val = long((high - low) * flat01() + low);
	RASSERT(val >= low && val < high);
	return val;
}

/*===========================================================================*/
double RGen::gauss_mean_d(double mean, double variance) {
	return gauss01() * variance + mean;	
}

/*===========================================================================*/
long RGen::gauss_mean_l(long mean, double variance) {
	return long(rint(gauss01() * variance + mean));
}

/*===========================================================================*/
double RGen::gauss_mean_half_d(double mean, double variance) {
	return mean + variance * abs(gauss01());
}

/*===========================================================================*/
long RGen::gauss_mean_half_l(long mean, double variance) {
	return long(rint(mean + variance * abs(gauss01())));
}

/*===========================================================================*/
bool RGen::flip(double cut_off) {
	return (flat01() < cut_off) ? 1 : 0;
}

/*===========================================================================*/
void RGen::test() {
#ifdef ROB_DEBUG
	RGen gen;
	eps_equal_to<double> eq;

	gen.set_seed(0);
	MAP(x, 25)
		Rassert(eq(gen.flat_range_d(-1024, 20), L_data1[x]));

	gen.set_seed(1);
	MAP(x, 25)
		Rassert(eq(gen.flat_range_d(-10, 2048), L_data2[x]));

	gen.set_seed(3);
	MAP(x, 25)
		Rassert(eq(gen.flat_range_d(234, 22523), L_data3[x]));

	gen.set_seed(51);
	MAP(x, 25)
		Rassert(eq(gen.flat_range_d(-1024, 2048), L_data4[x]));

	gen.set_seed(1024);
	MAP(x, 25)
		Rassert(eq(gen.flat_range_d(-51231, -123), L_data5[x]));

	gen.set_seed(0);
	MAP(x, 25)
		Rassert(eq(gen.gauss_mean_d(-1024, 2048), L_data6[x]));

	gen.set_seed(1);
	MAP(x, 25)
		Rassert(eq(gen.gauss_mean_d(-1024, 2048), L_data7[x]));

	gen.set_seed(3);
	MAP(x, 25)
		Rassert(eq(gen.gauss_mean_d(-1024, 2048), L_data8[x]));

	gen.set_seed(51);
	MAP(x, 25)
		Rassert(eq(gen.gauss_mean_d(-1024, 2048), L_data9[x]));

	gen.set_seed(1024);
	MAP(x, 25)
		Rassert(eq(gen.gauss_mean_d(-1024, 2048), L_data10[x]));
#endif

#if 0
	long cnt = 0;
	double var_sum = 0.0;

	while (1) {
		double x = RGen::gen.flat_range_d(0.0, 1.0);
		var_sum += (0.5 - x) * (0.5 - x);
		cnt++;

		cout << var_sum / cnt << "\t" << pow(var_sum / cnt, 0.5) << "\n";
	}
#endif

#if 0
	long a = 0, b = 0, c = 0, d = 0, e = 0, f = 0, g = 0,
	  h = 0, i = 0, j = 0;

	double total = 0.0;

	long ticker = 0;

	while (1) {		
		double fred = RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0) +
		  RGen::gen.flat_range_d(0.0, 1.0);

		if (fred < 1.0)
			a++;
		else if (fred < 2.0)
			b++;
		else if (fred < 3.0)
			c++;
		else if (fred < 4.0)
			d++;
		else if (fred < 5.0)
			e++;
		else if (fred < 6.0)
			f++;
		else if (fred < 7.0)
			g++;
		else if (fred < 8.0)
			h++;
		else if (fred < 9.0)
			i++;
		else
			j++;

		total += 1.0;

		if (ticker == 1000) {
			cout << setw(10) << a / total << "\t" <<
			  setw(10) << b / total << "\t" <<
			  setw(10) << c / total << "\t" <<
			  setw(10) << d / total << "\t" <<
			  setw(10) << e / total << "\t" <<
			  setw(10) << f / total << "\t" <<
			  setw(10) << g / total << "\t" <<
			  setw(10) << h / total << "\t" <<
			  setw(10) << i / total << "\t" <<
			  setw(10) << j / total << "\n";	
		}
		ticker++;
	}
#endif
}

/*############################################################################
Robert Dick (dickrp@wckn.dorm.clarkson.edu)

This is an implementation of the universal random number generator
proposed by G. Marsaglia and A. Zaman in
	G. Marsaglia, A. Zaman, W. Tseng
	Statistics and Probability Letters
	1990, volume 9, number 1, page 35.

The generation of i, j, k, and l from one seed is done using the
method proposed in
	F. James
	Computer Physics Communications
	1990, volume 60, page 329

Portable to machines with 32 bit floating point numbers with 24 bit
mantissas.

Gaussian RVs are generated using the Box-Muller transformation.  This is not
an approximation like the 12 sum approach.  True Gaussian RVs are generated.
############################################################################*/

#ifdef ROB_DEBUG
static double L_data1[] = {-663.698378086090087890625,
-940.3310453891754150390625, -837.9133317470550537109375,
-333.8126888275146484375, -682.1211907863616943359375,
-185.2112176418304443359375, -810.5848662853240966796875,
-651.8280947208404541015625, -812.0734043121337890625,
-477.9828331470489501953125, -218.960601806640625,
-744.0166442394256591796875, -352.9398517608642578125,
-125.90672016143798828125, -524.28955173492431640625,
8.8112294673919677734375, -13.404956340789794921875,
-869.47256565093994140625, -974.9584560394287109375,
-143.6278584003448486328125, -371.2813956737518310546875,
-320.4691753387451171875, -89.3641414642333984375,
-650.750132083892822265625, -228.1544911861419677734375};

static double L_data2[] = {1754.53165352344512939453125,
1769.16059815883636474609375, 1965.28263950347900390625,
1810.5986435413360595703125, 150.775852680206298828125,
535.46920764446258544921875, 830.266030788421630859375,
255.2019693851470947265625, 730.290985584259033203125,
192.70382750034332275390625, 606.46285355091094970703125,
640.9265804290771484375, 1462.026237964630126953125,
1778.5532839298248291015625, 390.6486909389495849609375,
1442.083141326904296875, 1974.567501544952392578125,
359.46089184284210205078125, 864.94491446018218994140625,
1869.9922225475311279296875, 191.3031003475189208984375,
259.1586954593658447265625, 536.35534942150115966796875,
1721.85407149791717529296875, 478.14304745197296142578125};

static double L_data3[] = {2597.885611474514007568359375,
12775.62513840198516845703125, 13550.3827698230743408203125,
10313.40321826934814453125, 2758.5697357654571533203125,
20858.741108596324920654296875, 22394.5951187610626220703125,
19525.97119152545928955078125, 21744.005693018436431884765625,
1827.472266376018524169921875, 17785.5066096782684326171875,
3210.100508034229278564453125, 21942.26058709621429443359375,
11482.721687376499176025390625, 600.000144302845001220703125,
17245.514476835727691650390625, 4723.729574501514434814453125,
7544.55705249309539794921875, 7543.1302134990692138671875,
15253.35629355907440185546875, 13249.350298225879669189453125,
3235.51391875743865966796875, 19345.32859218120574951171875,
21761.938162982463836669921875, 7833.252813756465911865234375};

static double L_data4[] = {1867.5718994140625, 1813.5184326171875,
1637.95635986328125, -688.979248046875, 1938.07080078125,
1960.90899658203125, 1614.23541259765625, 522.44927978515625,
94.67974853515625, 312.18359375, 1346.67803955078125,
744.0330810546875, 741.824462890625, -246.46881103515625,
470.40740966796875, -121.75970458984375, 1646.76025390625,
1428.157958984375, -526.28289794921875, -144.41278076171875,
822.48443603515625, -265.18109130859375, 432.09405517578125,
1602.470703125, 1229.708251953125};

static double L_data5[] = {-41646.452698230743408203125,
-2495.97447967529296875, -42056.17048358917236328125,
-14515.1376192569732666015625, -38773.7887251377105712890625,
-36808.29301929473876953125, -633.707867145538330078125,
-29427.1798074245452880859375, -40183.90599918365478515625,
-11897.879366397857666015625, -30175.0035645961761474609375,
-47370.7400286197662353515625, -21643.73297882080078125,
-25682.964604854583740234375, -39242.8052837848663330078125,
-37464.1710827350616455078125, -28332.4006519317626953125,
-2010.550688266754150390625, -9661.6643199920654296875,
-46933.3986289501190185546875, -21189.3354899883270263671875,
-18389.7546808719635009765625, -12601.3219549655914306640625,
-9903.532397747039794921875, -42231.227675914764404296875};

static double L_data6[] = {
1592.566360508824800490401685237884521484375,
417.52562575950969403493218123912811279296875,
-3039.8221022672159961075522005558013916015625,
-4249.441588440231498680077493190765380859375,
-15.717482918295672789099626243114471435546875,
-3913.3009218780998708098195493221282958984375,
-3287.52907029914103986811824142932891845703125,
1838.5137228208286614972166717052459716796875,
-4643.2171299971014377661049365997314453125,
-1550.811440524069666935247369110584259033203125,
-1192.335190593763172728358767926692962646484375,
442.99012866509519881219603121280670166015625,
205.59435702844939442002214491367340087890625,
-2505.7197167737385825603269040584564208984375,
1456.44858630336648275260813534259796142578125,
-1191.281952525074984805542044341564178466796875,
-711.75211525679833357571624219417572021484375,
-605.3112132198593826615251600742340087890625,
1777.17071583037977688945829868316650390625,
-5243.838735847870339057408273220062255859375,
-1937.55229895086768010514788329601287841796875,
-2786.1828713486320339143276214599609375,
-1626.48927417795357541763223707675933837890625,
-272.154710990377452617394737899303436279296875,
213.361892380153449266799725592136383056640625};

static double L_data7[] = {
-1887.5305490820492195780389010906219482421875,
-275.3544577301503295529983006417751312255859375,
-1878.4656862408692177268676459789276123046875,
-584.8345914582396289915777742862701416015625,
-1412.922601689368775623734109103679656982421875,
-1460.60425127265261835418641567230224609375,
3579.906437894338523619808256626129150390625,
866.70862755377811481594108045101165771484375,
960.816308326226817371207289397716522216796875,
1361.48157509038037460413761436939239501953125,
674.935327408451257724664174020290374755859375,
-2310.5591059561002111877314746379852294921875,
1884.12039614613831872702576220035552978515625,
116.575769333112475578673183917999267578125,
-2252.8473347147928507183678448200225830078125,
-2044.6998184731419314630329608917236328125,
-4585.647312925661026383750140666961669921875,
-787.44502329200577150913886725902557373046875,
-525.1710460018284720717929303646087646484375,
1268.688961836678572581149637699127197265625,
-2409.13425579896056660800240933895111083984375,
1982.99570686662036678171716630458831787109375,
2210.00538277972600553766824305057525634765625,
790.081300059943487212876789271831512451171875,
-3822.97329395677797947428189218044281005859375};

static double L_data8[] = {
-5030.3115235382492755888961255550384521484375,
-2688.84653021455642374348826706409454345703125,
-3009.697882437661974108777940273284912109375,
-409.214705076839663888677023351192474365234375,
2788.575430288614370510913431644439697265625,
-2956.5751480111312048393301665782928466796875,
-877.8554792564171975755016319453716278076171875,
-1188.6436054504765706951729953289031982421875,
-531.909458685519211940118111670017242431640625,
-786.782251486812810981064103543758392333984375,
-77.901885714993341025547124445438385009765625,
29.273714868751767426147125661373138427734375,
-1494.395253346550589412800036370754241943359375,
-1037.82405001714369063847698271274566650390625,
-536.7083986130850234985700808465480804443359375,
-6874.90927377235857420600950717926025390625,
-2749.4837898354589924565516412258148193359375,
2210.7765144254835831816308200359344482421875,
-2432.1463793026268831454217433929443359375,
-3738.8041674493788377731107175350189208984375,
384.209354094183026973041705787181854248046875,
566.50749906566261415719054639339447021484375,
85.858644701249659192399121820926666259765625,
-1265.831337780737840148503892123699188232421875,
173.022591595745097947656176984310150146484375};

static double L_data9[] = {
1731.65651146227537537924945354461669921875,
-391.8217291972742941652541048824787139892578125,
-1352.787047652479714088258333504199981689453125,
-175.1781725343572588826646097004413604736328125,
-330.230361608736757261794991791248321533203125,
-479.862602678119174015591852366924285888671875,
-1121.96437280484724396956153213977813720703125,
-2153.7426224386554167722351849079132080078125,
-1048.148527770542614234727807343006134033203125,
-3695.27365458397571273962967097759246826171875,
132.84826776346881160861812531948089599609375,
-2335.4798313830224287812598049640655517578125,
-1697.7595815942631816142238676548004150390625,
-1066.011172335226319773937575519084930419921875,
1130.7855021512050370802171528339385986328125,
-1690.59653515293348391423933207988739013671875,
1342.5316753141742083244025707244873046875,
-700.650464792049433526699431240558624267578125,
-2058.19630632336929920711554586887359619140625,
-1908.0627506758855815860442817211151123046875,
2782.06593361294699207064695656299591064453125,
-985.198313605670364268007688224315643310546875,
1042.07739740168881326098926365375518798828125,
509.461683474191886489279568195343017578125,
-3001.6726188174770868499763309955596923828125};

static double L_data10[] = {
2564.77449670403757409076206386089324951171875,
-2101.70892567042801601928658783435821533203125,
-1772.743955359010215033777058124542236328125,
-4745.1054414340178482234477996826171875,
-1715.486423508885536648449487984180450439453125,
2346.98688626937337176059372723102569580078125,
-1283.94689415873699545045383274555206298828125,
-894.8712146042433914772118441760540008544921875,
-583.4917604577066185811418108642101287841796875,
-4581.443049266103116679005324840545654296875,
1401.97259399942913660197518765926361083984375,
222.315591539162824119557626545429229736328125,
-3165.30450880865964791155420243740081787109375,
-1022.4298137527474636954138986766338348388671875,
-1447.36064174158809692016802728176116943359375,
2437.84976632488769610063172876834869384765625,
1501.6118818646491490653716027736663818359375,
-1620.83068463293102468014694750308990478515625,
112.897853939550486757070757448673248291015625,
-360.3981734558950620339601300656795501708984375,
-2343.1550658024525546352379024028778076171875,
-2672.382684302965571987442672252655029296875,
-472.183043203011266086832620203495025634765625,
-2453.55897462756502136471681296825408935546875,
-3215.13302117476587227429263293743133544921875};
#endif
