// Copyright 2000 by Robert Dick.
// All rights reserved.

#include <cstddef>
#include <set>
#include <algorithm>
#include <cmath>
#include <climits>
#include <iostream>
#include "RPair.h"
#include "Epsilon.h"
#include "RAlgo.h"
#include "RMath.h"
#include "RGen.h"

/*###########################################################################*/
static unsigned gcd(unsigned m, unsigned n) {
	unsigned t;

	do {
		t = n % m;
		n = m;
		m = t;
	} while (m);
	return n;
}

/*===========================================================================*/
static double lcm(double i, double j, double slack = 5.0) {
	eps_equal_to<double> eq(slack);

	if (eq(i, j))
		return i;


	unsigned i_mul = 1;
	unsigned j_mul = 2;

	double i_prod = i;
	double j_prod = j;

	do {
		if (i_prod < j_prod) {
			i_mul = unsigned(ceil(j_prod / i));
			i_prod = i_mul * i;
		} else {
			j_mul = unsigned(ceil(i_prod / j));
			j_prod = j_mul * j;
		}
	} while (! eq(i_prod, j_prod));

	return i_prod;
}

/*===========================================================================*/
static double gcd(double m, double n, double slack = 5) {
	return (m * n) / lcm(m, n, slack);
}

/*===========================================================================*/
static unsigned lcm(unsigned i, unsigned j) {
	return (i * j) / gcd(i, j);
}

/*===========================================================================*/
unsigned lcm(const RVector<unsigned> &num) {
	set<unsigned> srt(num.begin(), num.end());

	while (srt.size() > 1) {
		unsigned a = *srt.begin();
		srt.erase(srt.begin());

		unsigned b = *srt.begin();
		srt.erase(srt.begin());

		unsigned c = lcm(a, b);
		srt.insert(c);
	}

	return *srt.begin();
}

/*===========================================================================*/
double lcm(const RVector<double> &num, double slack) {
	set<double> srt(num.begin(), num.end());

	while (srt.size() > 1) {
		double a = *srt.begin();
		srt.erase(srt.begin());

		double b = *srt.begin();
		srt.erase(srt.begin());

		double c = lcm(a, b, slack);
		srt.insert(c);
	}

	return *srt.begin();
}

/*===========================================================================*/
unsigned gcd(const RVector<unsigned> &num) {
	set<unsigned> srt(num.begin(), num.end());

	while (srt.size() > 1) {
		unsigned a = *srt.begin();
		srt.erase(srt.begin());

		unsigned b = *srt.begin();
		srt.erase(srt.begin());

		unsigned c = gcd(a, b);
		srt.insert(c);
	}

	return *srt.begin();
}

/*===========================================================================*/
double gcd(const RVector<double> &num, double slack) {
	set<double> srt(num.begin(), num.end());

	while (srt.size() > 1) {
		double a = *srt.begin();
		srt.erase(srt.begin());

		double b = *srt.begin();
		srt.erase(srt.begin());

		double c = gcd(a, b, slack);
		srt.insert(c);
	}

	return *srt.begin();
}

/*###########################################################################*/
bool boltzmann_trial(double old_rank, double new_rank, double temperature,
RGen &rgen) {
	double cut = rgen.flat_range_d(0.0, 1.0);

	if (! temperature)
		return new_rank > old_rank;

	RASSERT(temperature);
	return 1.0 / (1.0 + exp((new_rank - old_rank) / temperature)) < cut;
}

/*###########################################################################*/
MathVec &MathVec::operator-=(const MathVec &a) {
	MAP(x, size())
		(*this)[x] -= a[x];
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::operator-=(const double &a) {
	MAP(x, size())
		(*this)[x] -= a;
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::operator+=(const MathVec &a) {
	MAP(x, size())
		(*this)[x] += a[x];
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::operator+=(const double &a) {
	MAP(x, size())
		(*this)[x] += a;
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::operator/=(const MathVec &a) {
	MAP(x, size())
		(*this)[x] /= a[x];
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::operator/=(const double &a) {
	MAP(x, size())
		(*this)[x] /= a;
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::flat_mul(const MathVec &a) {
	MAP(x, size())
		(*this)[x] *= a[x];
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::flat_mul(const double &a) {
	MAP(x, size())
		(*this)[x] *= a;
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::random_dir(double magnitude, RGen &rg) {
	bool all_zero = true;
	MAP(x, size()) {
		(*this)[x] = rg.gauss_mean_d(0.0, 1.0);
		if ((*this)[x])
			all_zero = false;
	}

	if (all_zero) {
// Do it the slow but certain way.
		MAP(x, size()) {
			if (rg.flip())
				(*this)[x] = max(DBL_MIN, rg.gauss_mean_half_d(0.0, 1.0));
			else
				(*this)[x] = min(-DBL_MIN, -rg.gauss_mean_d(DBL_MIN, 1.0));
		}
	}

	normalize(magnitude);
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::normalize(double magnitude) {
	double m = magnitude / mag();
	MAP(x, size())
		(*this)[x] *= m;
	return *this;
}

/*===========================================================================*/
MathVec &MathVec::clip(double low, double high) {
	MAP(x, size())
		(*this)[x] = std::clip((*this)[x], low, high);
	return *this;
}

/*===========================================================================*/
double MathVec::mag() const {
	double accum = 0.0;
	MAP(x, size())
		accum += (*this)[x] * (*this)[x];
	return sqrt(accum);
}

/*===========================================================================*/
void MathVec::print_to(ostream &os) const {
	os << "[";
	print(*this, os, " ", "]");
}

/*===========================================================================*/
double operator*(const MathVec &a, const MathVec &b) {
	RASSERT(a.size() == b.size());
	double accum = 0.0;
	MAP(x, a.size())
		accum += a[x] * b[x];

	return accum;
}

/*===========================================================================*/
RVector<double> operator*(const RVector<MathVec> &a, const MathVec &b) {
	RVector<double> vec(a.size());
	MAP(x, a.size())
		vec[x] = a[x] * b;
	return vec;
}

/*===========================================================================*/
MathVec max_dimen(RVector<MathVec> &vec) {
	MathVec max_v(vec[0].size(), -HUGE_VAL);
	MAP2(x, vec.size(), y, vec[x].size())
		max_v[y] = max(max_v[y], vec[x][y]);
	return max_v;
}

/*===========================================================================*/
MathVec min_dimen(RVector<MathVec> &vec) {
	MathVec min_v(vec[0].size(), HUGE_VAL);
	MAP2(x, vec.size(), y, vec[x].size())
		min_v[y] = min(min_v[y], vec[x][y]);
	return min_v;
}

/*===========================================================================*/
RVector<MathVec> &normalize(RVector<MathVec> &vec) {
	if (vec.empty())
		return vec;

	MathVec max_v = max_dimen(vec);
	MathVec min_v = min_dimen(vec);

	MathVec half_range = (max_v - min_v) / 2.0;
	MathVec mhr = min_v + half_range;

	double max_mag = 0.0;
	MAP(x, vec.size())
		max_mag = max(max_mag, vec[x].mag());

	MAP(x, vec.size())
		vec[x] = (vec[x] - mhr) / max_mag;

	return vec;
}

/*###########################################################################*/
void RMath_test() {
	Rassert(boltzmann_trial(0.0, 1.0, 0.0, RGen::gen));
	Rassert(! boltzmann_trial(1.0, 0.0, 0.0, RGen::gen));
}




