// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef R_STRING_H_
#define R_STRING_H_

/*###########################################################################*/
#include <cstddef>
#include <utility>
#include <functional>
#include "RVector.h"
class ostream;

/*###########################################################################*/
/* Some of the methods are overridden even though they are not virtual.
This means that someone working with an RString will get the RString
methods but internal RVector operations will get RVector methods.  I
meant to do this.  It keeps a '\0' under end(). */

class RString {
public:
	typedef char value_type;
	typedef value_type *pointer;
	typedef const value_type *const_pointer;
	typedef value_type *iterator;
	typedef const value_type *const_iterator;
	typedef value_type &reference;
	typedef const value_type &const_reference;
	typedef size_t size_type;
	typedef ptrdiff_t difference_type;
	typedef reverse_iterator<const_iterator> const_reverse_iterator;
	typedef reverse_iterator<iterator> reverse_iterator;
	class Converter;

// Initialization is deep from char * for convenience.
	RString(const char *a = "") : data_(a, a + strlen(a) + 1) {}

	template <typename InputIterator>
	  RString(InputIterator first, InputIterator last) :
	  data_(first, last) {data_.push_back(0);}

	explicit RString(bool);
	explicit RString(char a);
	explicit RString(int);
	explicit RString(unsigned);
	explicit RString(long);
	explicit RString(unsigned long);
	explicit RString(double);

	RString(const RString &a) : data_(a.data_) {}

	void reserve(size_type n) {data_.reserve(n + 1);}

	RString &operator+=(const RString &);

// Reads a line from istream and puts it in the RString.
	bool append_line(class istream &is, char terminator = '\n');

// Breaks up the RString at chars in divider, removing those chars.
	RVector<RString> tokenize(const char *divider = " \t") const;

// Locates the next divider in the RString.
	pair<RString, const_iterator> token(const_iterator first,
	  const char *divider = " \t") const;

	pair<RString, iterator> token(const_iterator first,
	  const char *divider = " \t");

// Breaks off the tail of the RString at the divider.
	pair<RString, RString> cut_tail(const char *divider = "/") const;

// Allows convenient conversion to another data type.
	const Converter convert() const {return Converter(*this);}

	friend bool operator==(const RString &a, const RString &b)
	  {return a.data_ == b.data_;}

	friend bool operator<(const RString &a, const RString &b)
	  {return a.data_ < b.data_;}

// Case insensitive comparisons.
	struct insen_equal_to : public binary_function<RString, RString, bool>
	  {bool operator()(const RString &a, const RString &b) const;};

	struct insen_less : public binary_function<RString, RString, bool>
	  {bool operator()(const RString &a, const RString &b) const;};

	iterator begin() {return data_.begin();}
	const_iterator begin() const {return data_.begin();}
	reverse_iterator rbegin() {return data_.rbegin();}
	const_reverse_iterator rbegin() const {return data_.rbegin();}
	const char *c_str() const {return data_.begin();}

	iterator end() {return data_.end() - 1;}
	const_iterator end() const {return data_.end() - 1;}
	reverse_iterator rend() {return data_.rend() + 1;}
	const_reverse_iterator rend() const {return data_.rend() + 1;}

	size_type size() const {return data_.size() - 1;}
	size_type max_size() const {return data_.max_size() - 1;}
	size_type capacity() const {return data_.capacity() - 1;}
	bool empty() const {return begin() == end();}
	reference operator[](size_type n) {return data_[n];}
	const_reference operator[](size_type n) const {return data_[n];}
 
	reference front() {return data_.front();}
	const_reference front() const {return data_.front();}
 	reference back() {return *(end() - 1);}
	const_reference back() const {return *(end() - 1);}
	void push_back(const_reference a);

	void swap(RString &s) {data_.swap(s.data_);}

	iterator insert(iterator position, char x)
	  {return data_.insert(position, x);}

	iterator insert(iterator position) {return data_.insert(position);}

	template <typename InputIterator>
	  void insert(iterator position, InputIterator first, InputIterator last)
	  {data_.insert(position, first, last);}

	void insert(iterator pos, size_type n, char x) {data_.insert(pos, n, x);}
	void insert(iterator pos, int n, char x) {data_.insert(pos, n, x);}
	void insert (iterator pos, long n, char x) {data_.insert(pos, n, x);}

	void pop_back() {data_.pop_back(); data_.back() = 0;}
	iterator erase(iterator position) {return data_.erase(position);}

	iterator erase(iterator first, iterator last)
	  {return data_.erase(first, last);}

	void resize(size_type new_size, char x)
	  {data_.resize(new_size + 1, x); back() = 0;}

	void resize(size_type new_size) {resize(new_size, char());}
	void clear() {data_.erase(begin(), end());}
 
	static void test();

private:
	template <typename T> void init_from(T);

	class Converter {
	public:
		Converter(const RString &a) : ref_(a) {}
		operator bool() const;
		operator int() const;
		operator unsigned() const;
		operator long() const;
		operator unsigned long() const;
		operator float() const;
		operator double() const;

	private:
		template <typename T> T convert_to() const;
			const RString &ref_;
	};

		RVector<char> data_;
};

inline RString operator+(const RString &a, const RString &b)
  {return RString(a) += b;}

inline ostream &operator<<(ostream &os, const RString &a)
  {return os << a.begin();}

/*###########################################################################*/
#endif
