// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef R_VECTOR_H_
#define R_VECTOR_H_

/*###########################################################################*/

#ifdef ROB_DEBUG
#include <iostream>
#include <typeinfo>
#endif

#include <vector>
#include <utility>
#include "RStd.h"
#include "RPair.h"

/*###########################################################################*/
/* Modifies vector for bounds checking and multidimensional
arrays.  No speed hits when compiled with ROB_DEBUG not defined. */

template <typename T> class RVector : public vector<T> {
	typedef vector<T> super;
	typedef RVector<T> self;

public:
	RVector() {}

	RVector(size_type n, const T &value) : super(n, value) {}
	RVector(int n, const T &value) : super(n, value) {}
	RVector(long n, const T &value) : super(n, value) {}
	RVector(const vector<T> &x) : super(x) {}
	explicit RVector(size_type n) : super(n) {}

	template <typename InputIterator>
	  RVector(InputIterator first, InputIterator last) :
	  super(first, last) {}

	reference operator[](size_type x);
	const_reference operator[](size_type x) const;
};

/*===========================================================================*/
template <typename T> class RVector2 : public RVector<RVector<T> > {
	typedef RVector<RVector<T> > super;
	typedef RVector2<T> self;

public:
	RVector2() {}

	RVector2(size_type x, size_type y, const T &value) :
	  super(x, typename self::value_type(y, value)) {}
	RVector2(int x, int y, const T &value) :
	  super(x, typename self::value_type(y, value)) {}
	RVector2(long x, long y, const T &value) :
	  super(x, typename self::value_type(y, value)) {}
	RVector2(size_type x, size_type y) : super(x, RVector<T>(y)) {}

	RVector2(size_type x, const RVector<T> &value) :
	  super(x, typename self::value_type(value)) {}
	RVector2(int x, const_reference value) :
	  super(x, typename self::value_type(value)) {}
	RVector2(long x, const_reference value) :
	  super(x, typename self::value_type(value)) {}
	explicit RVector2(size_type x) : super(x) {}

	RVector2(const super &x) : super(x) {}

	template <typename InputIterator>
	  RVector2(InputIterator first, InputIterator last) :
	  super(first, last) {}

	template <typename Container>
	  void resize2(const Container &c, const T &value);

	template <typename Container>
	  void resize2(const Container &c);

	reference operator[](size_type x)
	  {return super::operator[](x);}

	const_reference operator[](size_type x) const
	  {return super::operator[](x);}	

	T &operator[](const HPair<size_type> &i)
	  {return (*this)[i.first][i.second];}

	const T &operator[](const HPair<size_type> &i) const
	  {return (*this)[i.first][i.second];}
};

/*===========================================================================*/
template <typename T> class RVector3 : public RVector2<RVector<T> > {
	typedef RVector2<RVector<T> > super;
	typedef RVector3<T> self;

public:
	RVector3() {}

	RVector3(size_type x, size_type y, size_type z, const T &value) :
	  super(x, typename self::value_type(y, RVector<T>(z, value))) {}
	RVector3(int x, int y, int z, const T &value) :
	  super(x, typename self::value_type(y, RVector<T>(z, value))) {}
	RVector3(long x, long y, long z, const T &value) :
	  super(x, typename self::value_type(y, RVector<T>(z, value))) {}
	RVector3(size_type x, size_type y, size_type z) :
	  super(x, typename self::value_type(y, RVector<T>(z))) {}

	RVector3(size_type x, size_type y, const RVector<T> &value) :
	  super(x, typename self::value_type(y, value)) {}
	RVector3(int x, int y, const RVector<T> &value) :
	  super(x, typename self::value_type(y, value)) {}
	RVector3(long x, long y, const RVector<T> &value) :
	  super(x, typename self::value_type(y, value)) {}
	RVector3(size_type x, size_type y) :
	  super(x, typename self::value_type(y)) {}

	RVector3(size_type x, const RVector2<T> &value) :
	  super(x, value) {}
	RVector3(int x, const RVector2<T> &value) :
	  super(x, value) {}
	RVector3(long x, const RVector2<T> &value) :
	  super(x, value) {}
	explicit RVector3(size_type x) : super(x) {}

	RVector3(const super &x) : super(x) {}

	template <typename InputIterator>
	  RVector3(InputIterator first, InputIterator last) :
	  super(first, last) {}

	template <typename Container>
	void resize3(const Container &c, const T &value);

	template <typename Container>
	void resize3(const Container &c);

	reference operator[](size_type x)
	  {return super::operator[](x);}

	const_reference operator[](size_type x) const
	  {return super::operator[](x);}	

// Swap the next two lines when ndevry has up-to-date compiler.
	RVector<T> &
//	typename self::value_type::reference
	  operator[](const HPair<size_type> &i)
	  {return (*this)[i.first][i.second];}

// Swap the next two lines when ndevry has up-to-date compiler.
	const RVector<T> &
//	typename self::value_type::const_reference
	  operator[](const HPair<size_type> &i) const
	  {return (*this)[i.first][i.second];}

	T &operator[](const HTriple<size_type> &i)
	  {return (*this)[i.first][i.second][i.third];}

	const T &operator[](const HTriple<size_type> &i) const
	  {return (*this)[i.first][i.second][i.third];}
};

/*###########################################################################*/
void RVector_test();

/*###########################################################################*/
#include "RVector.cct"
#endif
