// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

//
//    File: dagman.h
//
//	Author: David L Rhodes
// Version: v2.0  --  December 29, 1997
//
// Purpose: To provide template class for dag management and
//			associated memory management functions. Note that
//			dagman always takes ownership of data
//			

#ifndef _DAGMAN
#define _DAGMAN

#include "listman.h"		// use listman to help in managing

class NodeObj : public ListObj {
public:
	SubListman parent_list, child_list;
};

class ArcObj : public ListObj {
public:
	NodeObj *from, *to;
};

template<class NODETYPE, class ARCTYPE>
class Dagman
{
	// NODETYPE must derive from NodeObj, ARCTYPE from ArcObj
	// both must have no argument constructors
protected:
	// contained objects:
	Listman<NODETYPE>	nodelist;
	Listman<ARCTYPE>	arclist;

	ARCTYPE				*startarc;

	// PUBLIC FUNCTIONS
public:
	// Constructors / Destructors
				Dagman() {startarc = 0;};
	virtual		~Dagman();

	// data addition/removal:
	void		AddNode(NODETYPE *);
	void		AddArc(ARCTYPE *, NODETYPE *, NODETYPE *);
	void		RemoveNode(NODETYPE *node);
	void		RemoveArc(ARCTYPE *);

	// data traversal:
	void		SetStart(ARCTYPE *);	// set start other than 1st ARC

	ARCTYPE		*GetStart() const	{ return startarc;};

	int			NumNodes(void) const;
	int			NumArcs(void) const;
	NODETYPE	*NodeElement(int k) const;
	ARCTYPE		*ArcElement (int k) const;
	int			NumOutArcs(NODETYPE *) const;
	int			NumInArcs(NODETYPE *) const;
	NODETYPE	*ToNode(ARCTYPE *) const;
	NODETYPE	*FromNode(ARCTYPE *) const;

	NODETYPE	*FirstNode() const;
	NODETYPE	*NextNode(NODETYPE *) const;	

	ARCTYPE		*FirstArc() const;
	ARCTYPE		*NextArc(ARCTYPE *) const;	

	SubListObj	*FirstInArc(NODETYPE *) const;
	SubListObj	*NextInArc(NODETYPE *, SubListObj *) const;

	SubListObj	*FirstOutArc(NODETYPE *) const;
	SubListObj	*NextOutArc(NODETYPE *, SubListObj *) const;

};
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
int			Dagman<NODETYPE,ARCTYPE>::NumNodes(void) const
{ return nodelist.NumOnList();};
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
int			Dagman<NODETYPE,ARCTYPE>::NumArcs(void) const
{ return  arclist.NumOnList();};
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
NODETYPE	*Dagman<NODETYPE,ARCTYPE>::NodeElement(int k) const
{ return nodelist[k];};
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
ARCTYPE		*Dagman<NODETYPE,ARCTYPE>::ArcElement (int k) const
{ return arclist[k]; };
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
Dagman<NODETYPE,ARCTYPE>::~Dagman()
{
	// contained lists are deleted
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
void Dagman<NODETYPE,ARCTYPE>::AddNode(NODETYPE *n)
{
	nodelist.InsertBefore(0,n);	// end of list
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
void Dagman<NODETYPE,ARCTYPE>::AddArc(ARCTYPE *a, NODETYPE *from, NODETYPE *to)
{
	// check that from and to are in here!
	RASSERT( nodelist.OnList(from) && nodelist.OnList(to) );
	a->from = from;
	a->to = to;
	arclist.InsertBefore(0,a);

	// add into parent/child lists...
	a->to->parent_list.InsertAfter(0,a);
	if( from ) a->from->child_list.InsertAfter(0,a);

	if( arclist.NumOnList() == 1 )
		startarc = a;
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
NODETYPE *Dagman<NODETYPE,ARCTYPE>::FirstNode() const
{
	return nodelist.First();
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
NODETYPE *Dagman<NODETYPE,ARCTYPE>::NextNode(NODETYPE *n) const
{
	return nodelist.Next(n);
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
ARCTYPE *Dagman<NODETYPE,ARCTYPE>::FirstArc() const
{
	return arclist.First();
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
ARCTYPE *Dagman<NODETYPE,ARCTYPE>::NextArc(ARCTYPE *a) const
{
	return arclist.Next(a);
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
SubListObj *Dagman<NODETYPE,ARCTYPE>::FirstInArc(NODETYPE *n) const
{
	return n->parent_list.First();
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
SubListObj *Dagman<NODETYPE,ARCTYPE>::NextInArc(NODETYPE *n,
												SubListObj *a) const
{
	return n->parent_list.Next(a);
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
SubListObj *Dagman<NODETYPE,ARCTYPE>::FirstOutArc(NODETYPE *n) const
{
	return n->child_list.First();
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
SubListObj *Dagman<NODETYPE,ARCTYPE>::NextOutArc(NODETYPE *n,
												SubListObj *a) const
{
	return n->child_list.Next(a);
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
int  Dagman<NODETYPE,ARCTYPE>::NumOutArcs(NODETYPE *n) const
{
	RASSERT( nodelist.OnList(n) );
	return n->child_list.NumOnList();
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
int  Dagman<NODETYPE,ARCTYPE>::NumInArcs(NODETYPE *n) const
{
	RASSERT( nodelist.OnList(n) );
	return n->parent_list.NumOnList();
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
void Dagman<NODETYPE,ARCTYPE>::RemoveNode(NODETYPE *n)
{
	RASSERT( nodelist.OnList(n) );
	// removes node and all arcs to/from

	// first delete incoming/outgoing arcs ...
	ARCTYPE *iarc;
	SubListObj	*parent_list = n->parent_list;
	while( iarc = parent_list->First() )
		RemoveArc(iarc);

	SubListObj *child_list = n->child_list;
	while( iarc = child_list->First() )
		RemoveArc(iarc);

	// delete node ...
	nodelist.Remove(n);
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
void Dagman<NODETYPE,ARCTYPE>::RemoveArc(ARCTYPE *arc)
{
	NODETYPE *from = arc->from;
	if( from ) from->child_list.Remove(arc);// rm from child list of from
	NODETYPE *to   = arc->to;
	RASSERT( nodelist.OnList(to) );
	to->parent_list.Remove(arc);			// rm from parent list of to

	arclist.Remove(arc);	// delete arc.
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
void  Dagman<NODETYPE,ARCTYPE>::SetStart(ARCTYPE *a)
{
	RASSERT( arclist.OnList(a) );
	startarc = a;
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
NODETYPE *Dagman<NODETYPE,ARCTYPE>::ToNode(ARCTYPE *arc) const
{
	return (NODETYPE *)arc->to;
}
// --------------------------------------------------------
template<class NODETYPE, class ARCTYPE>
NODETYPE *Dagman<NODETYPE,ARCTYPE>::FromNode(ARCTYPE *arc) const
{
	return (NODETYPE *)arc->from;
}
// --------------------------------------------------------

// UTILITY DEFs (ITERATORS):

// Over nodes:

#define DAG_FOREACH_NODEC(ntype,n,dag,__k)									\
	{for(int __k=0; __k < (dag)->NumNodes(); __k++){						\
		ntype *n = (dag)->NodeElement(__k);
	
#define DAG_ENDEACH_NODEC	}}

#define DAG_FOREACH_NODE(ntype,n,dag)										\
	{for(ntype *n = (dag)->FirstNode(); n; n = (dag)->NextNode(n)){

#define DAG_ENDEACH_NODE	}}

// Over arcs:

#define DAG_FOREACH_ARCC(atype,a,dag,__k)									\
	{for(int __k=0; __k < (dag)->NumArcs(); __k++){							\
		atype *a = (dag)->ArcElement(__k);
	
#define DAG_ENDEACH_ARCC		}}

#define DAG_FOREACH_ARC(atype,a,dag)										\
	{for(atype *a = (dag)->FirstArc(); a; a = (dag)->NextArc(a) ){	

#define DAG_ENDEACH_ARC	}}

// Over outgoing/incoming arcs
// ... Beware side-effects! Beware Add/Remove during Iterating
//!

// OUTARCs :

#define DAG_FOREACH_OUTARC(atype,a,n,dag)									\
	{for(SubListObj *s = (dag)->FirstOutArc(n); s;							\
						s = (dag)->NextOutArc(n,s) ){						\
		atype *a = (atype *) s->data;

#define DAG_ENDEACH_OUTARC	}}

// INARCS:

#define DAG_FOREACH_INARC(atype,a,n,dag)									\
	{for(SubListObj *s = (dag)->FirstInArc(n); s;							\
						s = (dag)->NextInArc(n,s) ){						\
		atype *a = (atype *) s->data;

#define DAG_ENDEACH_INARC	}}

#endif _DAGMAN

