// Copyright 2000 by David Rhodes and Robert Dick.
// All rights reserved.

//
//    File: listman.h
//
//	Author: David L Rhodes
// Version: v2.0  --  December 29, 1997
//
// Purpose: To provide template class for intrusive list management and
//			associated memory management functions. Listman takes
//			`ownership' of the objects, while SubListman only manages
//          void pointers. The "P" version of each adds print out via <<
//

#ifndef _LISTMAN
#define _LISTMAN

#include <iostream.h>
#include "RStd.h"		// ONLY to get define for ASSERT

// -------------------------------------------------------
// ---Listman---------------------------------------------
// -------------------------------------------------------

class ListObj {				// base class for Intrusive List
public:
	ListObj	*next;
	ListObj	*prev;
};
// -------------------------------------------------------
template<class TYPE>
class Listman	
{
// TYPE must derive from ListObj and have default constructor
// implementation is double-linked list
protected:
	TYPE	*root;
	int		n_onlist;

	// PUBLIC FUNCTIONS
public:
	// Constructors / Destructors
	Listman();
	Listman(const Listman<TYPE> &);
	~Listman();
	
	// retreive information
	int		NumOnList(void)	const	{ return n_onlist; }
	int		OnList(TYPE *obj) const;
	TYPE	*operator[](int index) const; 
	inline TYPE	*First(void) const;
	inline TYPE	*Last (void) const;
	inline TYPE	*Next (TYPE *curobj) const;
	inline TYPE	*Prev (TYPE *curobj) const;
	
	// modify list
	void	InsertAfter (TYPE *listobj,TYPE *listaddition);
	void	InsertBefore(TYPE *listobj,TYPE *listaddition);
	void	Cut(TYPE *obj);
	void	Delete(TYPE *obj);
	void	Empty(void);
};
// -------------------------------------------------------
template<class TYPE> 
Listman<TYPE>::Listman()
{
	// create an empty list
	root = new TYPE;
	root->next = root->prev = root;
	n_onlist = 0;
}

/*===========================================================================*/
template<class T>
Listman<T>::Listman(const Listman<T> &src) :
	root(new T),
	n_onlist(0)
{
	root->next = root->prev = root;
	Rassert(! src.n_onlist);
}

// --------------------------------------------------------
template<class TYPE> 
Listman<TYPE>::~Listman()
{
	// destroy everything
	Empty();
	delete root;
}
// --------------------------------------------------------
template<class TYPE>
void Listman<TYPE>::Empty()
{
	// destroy everything
	TYPE *p = (TYPE *)root->next;
	while( p != root ) {
		TYPE *nxt = (TYPE *)p->next;
		delete p;
		p = nxt;
	}

	n_onlist = 0;
}
// --------------------------------------------------------
template<class TYPE>
int Listman<TYPE>::OnList(TYPE *obj) const
{
	TYPE *p = (TYPE *)root->next;
	while( p != root ) {
		if( obj == p ) return 1;
		p = (TYPE *)p->next;
	}

	return 0; // not found
}
// --------------------------------------------------------
template<class TYPE>
void Listman<TYPE>::Cut(TYPE *obj)
{
	ASSERT( OnList(obj) );
	obj->next->prev = (TYPE *)obj->prev;
	obj->prev->next = (TYPE *)obj->next;

	n_onlist--;
}
// --------------------------------------------------------
template<class TYPE>
void Listman<TYPE>::Delete(TYPE *obj)
{
	Cut(obj);
	delete obj;
}
// --------------------------------------------------------
template<class TYPE>
TYPE *Listman<TYPE>::operator[](int k) const
{
	// return k-th element ptr where k=0 implies first in list
	if( k >= n_onlist || k < 0 ) return 0;
	TYPE *p = (TYPE *)root->next;
	for(int j=1; j <= k; j++ ) {
		p = (TYPE *)p->next;
	}

	return p;
}
// --------------------------------------------------------
template<class TYPE>
inline TYPE *Listman<TYPE>::First(void) const
{
	return (((TYPE *)root->next == root) ? 0 : (TYPE *)root->next);
}
// --------------------------------------------------------
template<class TYPE>
inline TYPE *Listman<TYPE>::Last(void) const
{
	return ((TYPE *)root->prev == root ? 0 : (TYPE *)root->prev);
}
// --------------------------------------------------------
template<class TYPE>
inline TYPE *Listman<TYPE>::Next(TYPE *curobj) const
{
	return ((TYPE *)curobj->next == root ? 0 : (TYPE *)curobj->next);
}
// --------------------------------------------------------
template<class TYPE>
inline TYPE *Listman<TYPE>::Prev(TYPE *curobj) const
{
	return ((TYPE *)curobj->prev == root ? 0 : (TYPE *)curobj->prev);
}
// --------------------------------------------------------
template<class TYPE>
void Listman<TYPE>::InsertAfter(TYPE *listobj,TYPE *listaddition)
{
	// if listobj == NULL then the insert is at the head
	ASSERT( !listobj || OnList(listobj) );
	ASSERT( !OnList(listaddition) );	// will get ill-formed list then
	TYPE *before = (listobj == 0 ? root : listobj);
	TYPE *after  = (TYPE *)before->next;
	listaddition->next = after;
	listaddition->prev = before;
	before->next = (TYPE *)after->prev = (TYPE *)listaddition;
	
	n_onlist++;
}
// --------------------------------------------------------
template<class TYPE>
void Listman<TYPE>::InsertBefore(TYPE *listobj,TYPE *listaddition)
{
	// if listobj == NULL then the insert is at the tail
	RASSERT( !listobj || OnList(listobj) );
	RASSERT( !OnList(listaddition) );	// will get ill-formed list then
	TYPE *after  = (listobj == 0 ? root : listobj);
	TYPE *before = (TYPE *)after->prev;
	listaddition->next = after;
	listaddition->prev = before;
	before->next = (TYPE *)(after->prev = (TYPE *)listaddition);
	
	n_onlist++;
}
// -------------------------------------------------------
// ---SubListman------------------------------------------
// -------------------------------------------------------

class SubListObj : public ListObj {	// base class for Intrusive SubList
public:
	void		*data;
};
// -------------------------------------------------------
class SubListman	
{
// TYPE must derive from SubListObj and have default constructor
// implementation is double-linked list using hidden ptrs
protected:
	SubListObj	*root;
	int			n_onlist;

	// PUBLIC FUNCTIONS
public:
	// Constructors / Destructors
	SubListman();
	~SubListman();
	
	// retreive information
	int		NumOnList(void)	const	{ return n_onlist; }
	int		OnList(SubListObj *obj) const;
	SubListObj	*operator[](int index) const; 
	/*inline*/ SubListObj	*First(void) const;
	/*inline*/ SubListObj	*Last (void) const;
	/*inline*/ SubListObj	*Next (SubListObj *curobj) const;
	/*inline*/ SubListObj	*Prev (SubListObj *curobj) const;
	
	// modify list
	SubListObj	*InsertAfter (SubListObj *listobj,void *listaddition);
	SubListObj	*InsertBefore(SubListObj *listobj,void *listaddition);
	void	Remove(SubListObj *obj);	// Cut doesn't make sense for SubList
	void	Empty(void);
};

// -------------------------------------------------------
// ---PListman--------------------------------------------
// -------------------------------------------------------
class PListObj : public ListObj {	// base class for Intrusive List w/ Print
public:
	friend ostream& operator<<(ostream&s, PListObj& obj);
};
// -------------------------------------------------------
template<class TYPE>
class PListman : public Listman<TYPE>
{
private:
	PListman(void);	// hide
protected:
	char *sep;
public:
	PListman(const char *sepinit = "\n");
	friend ostream& operator<< <>(ostream&s, PListman& obj);
};

// --------------------------------------------------------
// --------------------------------------------------------
// --------------------------------------------------------
// UTILITY DEF: (this works for any of the lists)

#define LM_FOREACH(ltype,e,list)									\
	{for(ltype *e = (list)->First(); e; e = (list)->Next(e)){

#define LM_ENDEACH	}}

#endif _LISTMAN

